@@ -17,7 +17,7 @@
 # along with this program.  If not, see <https://www.gnu.org/licenses/>.
 
 
-from ..core import *
+from fcsugar import *
 import numpy as np
 from scipy import signal as scipy_signal
 from ..containers import TransmissionContainer
